package experiment;

import experiment.property.Position;

import java.text.SimpleDateFormat;
import java.util.Date;

import walker.utilities.Output;
import static walker.utilities.Output.print;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 9, 2010
 *
 * Corresponds to the setup tag of a wiseml file which contains the setup information of the
 * experiment.
 */
public class Setup {


    public static enum Datatype {
        decimal, integer

    }


    public static enum Unit {
        meters, centimeters, radians, degrees, seconds, milliseconds, nanoseconds, dBm, kelvin, percentage, lux, raw, other

    }

    /**
     * Information about the relative system of coordinates
     */
    private static Position position;

    /**
     * Information regarding the time used in the experiment.
     * The tags start and end define the start and end time of the experiment (in the ISO 8601 format).
     * All the time information in the document will be stored as a relative value from this time in a
     * unit specified by the user. The end tag can be replaced with duration, if this is more convenient.
     *
     * For easier manipulation all application time values are stored in milliseconds. When converting
     * from and to wiseml timestamps the translateTime  method should be used for translation.
     *
     * TODO: To allow greater flexibility and precision the application's internal time unit should migrate to nanoseconds.
     */
    private static Date start, end;

    private static long duration;

    private static Unit unit;


    /**
     * All different interpolation options
     */
    public static enum Interpolation {
        none, linear, polynomial, cubic, spline

    }

    private static Interpolation interpolation;


    /**
     * Unit type value of the xyz values for the position of the node.
     * Visualizator works only with xyz values.
     */
    public static enum CoordinateType {
        cartesian, global_positioning, XYZ

    }

    private static CoordinateType coordType;

    /**
     * Contains a textual description of the experiment (and any other information such as environmental info, etc)
     */
    private static String description;

    // Setting the time format
    public static SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss"); //check date (Z)

    /*

        Setters

    */


    /**
     * Set the origin info of the experiment.
     * @param position the (0,0,0) point in the local system of coordinates and the angles of rotation.
     */
    public static void setOrigin(Position position) {
        Setup.position = position;
    }

    /**
     * Set the time information of the experiment.
     *
     * @param start the starting time of the experiment in a Date object.
     * @param end the enging time of the experiment in a Date object.
     * @param unit the time unit
     */
    public static void setTime(Date start, Date end, Unit unit) {
        Setup.start = start;
        Setup.end = end;
        Setup.duration = end.getTime() - start.getTime();
        Setup.unit = unit;
        print("timeinfo: experiment started at " + start + " ended at " + end + "and time unit is " + unit);
    }

    /**
     * Set the time information of the experiment.
     *
     * @param start the starting time of the experiment in a Date object.
     * @param duration a long containing the duration of the experiment in milliseconds.
     * @param unit the time unit
     */
    public static void setTime(Date start, long duration, Unit unit) {
        Setup.start = start;
        Setup.duration = duration;
        Setup.end = new Date(Setup.start.getTime() + Setup.duration);
        Setup.unit = unit;
        print("timeinfo: experiment started at " + start + " ended at " + end + "and time unit is " + unit);
    }

    /**
     * Set the interpolation that is going to be used in this experiment.
     * @param interpolation type of interpolation that should be used for the description of mobility.
     */
    public static void setInterpolation(Interpolation interpolation) {
        Setup.interpolation = interpolation;
    }

    /**
     * Set the coordType value for this experiment.
     * Only works for xyz coordinates.
     *
     * @param s the String representation of the coordinate type.
     */
    public static void setCoordType(String s) {
        CoordinateType c = null;
        try {
            c = CoordinateType.valueOf(s);
        } catch (IllegalArgumentException e){
            print("coordinate type '" + s + "' not supported");
            c = CoordinateType.cartesian;
        }
        if (c!=CoordinateType.cartesian || c!=CoordinateType.XYZ){
            print("Visualizer only works for 'cartesian' coordinateType");
        }
        Setup.coordType = CoordinateType.cartesian;
    }

    public static void setCoordType(CoordinateType coordinateType) {
        Setup.coordType = coordinateType;

    }

    /**
     * Set the experiment's description.
     * @param description of the experiment (and any other information such as environmental info, etc).
     */
    public static void setDescription(String description) {
        Setup.description = description;
    }


    /*

        Getters

    */

    /**
     * Safely returns the Unit type that matches the input. In case that there is no match returns Unit.other
     *
     * @param s String input
     * @return the Unit type
     */
    public static Unit toUnit(String s){
        Unit unit = null;
        try {
            unit = Unit.valueOf(s);
        } catch (IllegalArgumentException e){
            print("not supported unit type '" + s + "'");
            unit = Setup.Unit.other;
        }
        return unit;
    }

    /**
     * Returns the center of the relative system of coordinates
     *
     * @return a Position object containing the center of the plane
     */
    public static Position getPosition() {
        return position;
    }

    /**
     * Returns the experiment's starting time
     * @return a Date object containing the experiment's starting time
     */
    public static Date getStart() {
        return start;
    }

    /**
     * Returns the experiment's ending time
     * @return a Date object containing the experiment's ending time
     */
    public static Date getEnd() {
        return end;
    }

    /**
     * Returns the experiment's duration
     *
     * @return the experiment's duration in milliseconds
     */
    public static long getDuration() {
        return duration;
    }

    /**
     * Returns the type of interpolation that is used for the description of mobility
     * @return an Interpolation enum value
     */
    public static Interpolation getInterpolation() {
        return interpolation;
    }

    /**
     * Returns the unit type value of the xyz values
     * @return an CoordinateType enum value
     */
    public static CoordinateType getCoordType() {
        return coordType;
    }

    /**
     * Returns a textual description of the experiment
     * @return a String containing a description of the experiment
     */
    public static String getDescription() {
        return description;
    }

    /**
     * Adjusts the timestamp values to milliseconds according to the time unit. Helps the tranlation of wiseml
     * timestamps to the application's time unit and vise versa. To translate from experiment to application unit,
     * multiply with the method's result, else divide.
     *
     * @return the time unit translation factor.
     */
    public static double translateTime() {

        if (unit == Unit.milliseconds){
            return 1;
        } else if (unit == Unit.seconds) {
            return 1000;
        } else if (unit == Unit.nanoseconds) {
            return 0.001;
        } else {
            return 1;
        }

    }

}
