package experiment.entity;

import experiment.ExperimentState;
import experiment.property.Capability;
import visual.entity.Entity3d;
import walker.Weyes;

import java.util.HashMap;
import java.util.Collection;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 11, 2010
 */
public class Entity {

    // An entity can be enabled or disabled
    protected volatile boolean enabled = true;
    // Every entity object belongs to an experiment state
    protected ExperimentState state;
    // Stores the 3d shape that represents this Entity.
    public transient Entity3d entity3d = null;
    // A monitored entity is being described by the FrontEnd gui.
    private volatile boolean monitored = false;
    // Is used to distinguish between Node and Link instances.
    protected boolean isNode;
    // Stores capabilities in a HashMap with the capability name as key.
    protected HashMap<String, Capability> capabilityMap = new HashMap<String, Capability>(1);
    // Locks the capabilityMap to prevent concurrentModifications.
    public final Object lock = new Object();

    /**
     * Default Constructor
     *
     * @param state an experiment state.
     */
    public Entity(ExperimentState state) {
        this.state = state;
    }

    /**
     * Find out if the entity is enabled
     *
     * @return true if the entity is enabled
     */
    public boolean isEnabled() {
        return enabled;
    }

    /**
     * Set the entity's status.
     *
     * @param status true to enable it and false to disable it
     */
    public void setStatus(boolean status) {
        enabled = status;   //todo: update the frontEnd, 3d
    }

    /**
     * @return the object's state
     */
    public ExperimentState getState() {
        return state;
    }

    /**
     * Set the entity's state
     *
     * @param state the entity's state
     */
    public void setState(ExperimentState state) {
        this.state = state;
    }

     /**
     * Update a capability if it's already known else create it.
     * It is called during experiment initilization i.e. during setup parsing.
      *
     * @param c the new capability
     */
     public void setCapability(Capability c){
         synchronized(lock){
             if (capabilityMap.containsKey(c.getName())){
                 capabilityMap.get(c.getName()).update(c.getValue());
             } else {
                 capabilityMap.put(c.getName(),c);
             }
         }
     }

    /**
     * Update a capability if it's already known else create it.
     * It is usually called during experiment visualization i.e. during trace parsing.
     *
     * @param name the name of the capability that has to be updated.
     * @param value the new value.
     */
    public void setCapability(String name, String value){
        synchronized(lock){
            // Index the capability
            Capability cap = capabilityMap.get(name);

            // if it is found update its value.
            if (cap != null) {
                cap.update(value);
                // if it isn't there create it.
            } else {
                capabilityMap.put(name, new Capability(name, null, null, value));
            }
        }
        // Signal an update.
        updated();
    }

    /**
     * @return a Collection of all the entity's Capabilities.
     */
    public Collection<Capability> getCapabilitySet() {
        synchronized(lock){
            return capabilityMap.values();
        }
    }


    /**
     * Return a capability of the entity.
     *
     * @param name the name of the capability
     * @return the capability if it exists else null
     */
    public Capability getCapability(String name){
        synchronized(lock){
            return capabilityMap.get(name);
        }
    }

    /**
     * @return the Entity3d object that is the visual 3d representation of this object.
     */
    public Entity3d getEntity3d() {
        return entity3d;
    }

    /**
     * Associates an Entity object with its visual representation.
     * 
     * @param entity3d contains the 3d representation of this object.
     */
    public void setEntity3d(Entity3d entity3d) {
        this.entity3d = entity3d;
    }

    /**
     * Signals a change in this object since it has last been monitored.
     * Is called in the process of notifying the FrontEnd that the monitored Entity was updated.
     */
    public void updated(){
        if (monitored){
            synchronized (Weyes.fe.lock){
                Weyes.fe.lock.notifyAll();
            }
        }
    }

    /**
     * A monitored entity is being displayed by the FrontEnd gui.
     *
     * @param bool true for monitored, false for not.
     */
    public void setMonitored(boolean bool){
        monitored = bool;
    }

    /**
     * Used tp distinguish between Node and Links.
     *
     * @return true if the instance is a Node.
     */
    public boolean isNode() {
        return isNode;
    }
}
