package experiment.entity;

import experiment.property.Rssi;
import experiment.property.Capability;
import experiment.ExperimentState;

import java.util.HashMap;
import walker.utilities.SerialClone;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 10, 2010
 *
 * Represent's an experiment link and stores all the link's data.
 */
public class Link extends Entity {

    /**
     * Static default variables
     */
    public static boolean defaultIsEncrypted = false;
    public static boolean defaultIsVirtual = false;
    public static Rssi defaultRssi = null; //todo what is default?

    // Stores default capabilities in a HashMap with the capability name as key.
    private static HashMap<String, Capability> defaultCapabilityMap = new HashMap<String, Capability>();

    /**
     * Non-Static Variables
     */
    private Node source;
    private Node target;
    private boolean isEncrypted;
    private boolean isVirtual;
    private Rssi rssi;

    /**
     * Full Constructor. Put null to all not available and not required attributes.
     *
     * @param state the object that contains all the data for a given experiment state
     * @param source the source Node
     * @param target the target Node
     * @param encrypted is a Boolean value in a String indicating if the link is encrypted or not.
     * @param virtual is a Boolean value in a String indicating if the link is virtual or not.
     * @param rssi specifies the RSSI value on the link
     */
    public Link(ExperimentState state, Node source, Node target, String encrypted, String virtual, Rssi rssi) {

        // Set the Link's state
        super(state);

        this.source = source;
        this.target = target;

        initialize();

        if (encrypted!=null){
            this.isEncrypted = Boolean.parseBoolean(encrypted);
        }
        if (virtual!=null){
            this.isVirtual = Boolean.parseBoolean(virtual);
        }
        if (rssi!=null){
            this.rssi = rssi;
        }

        isNode = false;

        // Initialize the capability map with the default values
        capabilityMap = SerialClone.clone(defaultCapabilityMap);

        // Notify source Node3d for the new linkage
        source.addLink(this, target.getId());
        // add the link to the current experiment state

        // add the link to its state.
        getState().add(this);

        updated();
    }

    /**
     * Set's all Link variables to the default value
     */
    private void initialize() {
        this.isEncrypted = defaultIsEncrypted;
        this.isVirtual = defaultIsVirtual;
        this.rssi = defaultRssi;
    }

     /*
     *
     * Static Methods
     *
     */

    /**
     * Is called during experiment configurarion to set the default link capabilities
     * @param dC the capability to be set
     */
    public static void addDefaultCapability(Capability dC){
        if (!defaultCapabilityMap.containsKey(dC.getName())){
            defaultCapabilityMap.put(dC.getName(), dC);
        }
    }

    /**
     *
     * Non-Static non-initialization methods
     *
     */

    /**
     * Set encryption mode.
     *
     * @param encrypted true if the link is enscrypted
     */
    public synchronized void setEncrypted(boolean encrypted) {
        if (isEncrypted!=encrypted){
            updated();
        }
        isEncrypted = encrypted;
    }

    /**
     * Set the Link's rssi.
     *
     * @param rssi the new rssi object
     */
    public synchronized void setRssi(Rssi rssi) {
        this.rssi = rssi;
        updated();
    }

    /**
     * @return the source Node of the link
     */
    public synchronized Node getSource() {
        return source;
    }

    /**
     * @return the target Node of the link
     */
    public synchronized Node getTarget() {
        return target;
    }

    /**
     * @return true if the Link is encrypted
     */
    public synchronized boolean isEncrypted() {
        return isEncrypted;
    }

    /**
     * @return true if this is a virtual Link.
     */
    public boolean isVirtual() {
        return isVirtual;
    }

    /**
     * @return the Rssi of the Link
     */
    public synchronized Rssi getRssi() {
        return rssi;
    }

}
