package visual;

import experiment.Setup;
import walker.Persistor;
import walker.Weyes;
import walker.utilities.Rendering;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.net.URL;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Apr 9, 2010
 *
 * Simple Gui that simulates the way that a player works.
 * It allows playing, stopping, pausing, seeking and speed calibration of the visualization.
 */
public class Player implements MouseListener{
    private static final int SLIDER_MAX = 10000;
    private JPanel container;
    private JSlider slider;
    private JButton playButton;
    private JButton pauseButton;
    private JButton stopButton;
//    private JProgressBar progress;
    private JButton previousButton;
    private JButton nextButton;
    private JButton snapButton;
    private JButton speedButton;
    private JLabel logo;
    private JButton cameraButton;
    private int cameraMode = 1;

    public volatile boolean run = false, reset = false;
    public volatile double seeked = -1;
    public volatile long seekedTimestamp = -1;

    public Player() {
        container.setMinimumSize(new Dimension(Rendering.getScreenWidth(), 100));
        // Set max values.
        slider.setMaximum(SLIDER_MAX);
//        progress.setMaximum(SLIDER_MAX);

        // Create buttons.
        makeButtonsJar();

        /*
         * GUI Listeners:
         */

        /**
         * Button Listener.
         */
        ActionListener listener = new ActionListener() {
            public void actionPerformed(ActionEvent actionEvent) {
                // Pause visualization.
                if (actionEvent.getSource().equals(pauseButton)){
                    run = false;
                    seeked = -1;
                    System.out.println("PAUSED");

                // Play visualization.
                } else if (actionEvent.getSource().equals(playButton)){
                    run = true;
                    seeked = -1;
                    // Notify the Crawler.
                    synchronized (Crawler.getLock()) {
                        Crawler.getLock().notifyAll();
                    }
                    System.out.println("PLAYING");

                // Stop visualization.
                } else if (actionEvent.getSource().equals(stopButton)){
                    run = false;
                    reset = true;
                    seeked = 0;
                    // Notify the Crawler.
                    synchronized (Crawler.getLock()) {
                        Crawler.getLock().notifyAll();
                    }
                    System.out.println("STOPPED");

                // Go one step back.
                } else if (actionEvent.getSource().equals(previousButton)){
                    System.out.println("PREVIOUS not implemented yet");

                // Go one step forward.
                } else if (actionEvent.getSource().equals(nextButton)){
                    System.out.println("NEXT not implemented yet");

                // Keep a snapshot of the ExperimentState.
                } else if (actionEvent.getSource().equals(snapButton)){
                    Crawler.getExperimentState().setByteOffset();
                    Persistor.store(Crawler.getExperimentState());
                    System.out.println("new SNAPSHOT");

                } else if (actionEvent.getSource().equals(speedButton)){
                    new SpeedConfig();

                } else if (actionEvent.getSource().equals(cameraButton)){
                    Weyes.crawler.engine3d.resetCamera(cameraMode);
                    if (cameraMode == 1) {
                        cameraMode = 2;
                    } else {
                        cameraMode = 1;
                    }
                }
            }
        };

        /**
         * Seek listener.
         */
        slider.addMouseListener(this);

        pauseButton.addActionListener(listener);
        stopButton.addActionListener(listener);
        playButton.addActionListener(listener);
        previousButton.addActionListener(listener);
        nextButton.addActionListener(listener);
        snapButton.addActionListener(listener);
        speedButton.addActionListener(listener);
        cameraButton.addActionListener(listener);

    }

    public void mouseReleased(MouseEvent e){

        int sliderValue = slider.getValue();

        // calculate the seeked timestamp.
        if (sliderValue >= 0 && sliderValue < SLIDER_MAX){

            // calculate the seeked time. It holds a relative time value.
            seeked = ((Setup.getDuration()*sliderValue)/SLIDER_MAX);// + Setup.getStart().getTime();
            run = false;
            reset = true;

            // update frontEnd's time
            Weyes.fe.setTime(seeked);

            // Notify the Crawler.
            synchronized (Crawler.getLock()) {
                Crawler.getLock().notifyAll();
            }
        }
    }

    public void mouseDragged(MouseEvent e){}

    public void mouseClicked(MouseEvent mouseEvent) {}

    public void mousePressed(MouseEvent mouseEvent) {}

    public void mouseEntered(MouseEvent mouseEvent) {}

    public void mouseExited(MouseEvent mouseEvent) {}

    /**
     * Changes the slider's value to match the visualization time. The Slider is split SLIDER_MAX times.
     *
     * @param time the current visualization time in milliseconds.
     */
    public void setSlider(double time){
        slider.setValue((int) (time*SLIDER_MAX/Setup.getDuration()));
    }

    /**
     * @return the JPanel that contains this gui.
     */
    public JPanel getContainer(){
            return container;
    }

//    public void incrementProgress(){
//        progress.setValue(progress.getValue()+1);
//    }
//
//    public void incrementProgress(double time) {
//        int progr = (int) (time*SLIDER_MAX/Setup.getDuration());
//        if (progr > progress.getValue()){
//            progress.setValue(progr);
//        }
//    }

    /**
     * Make the buttons of the gui look nicer.
     */
    private void makeButtonsJar(){

        final ClassLoader loader = this.getClass().getClassLoader();
        final URL play = loader.getResource("ico/play.png");
        final URL playRollover = loader.getResource("ico/playRollover.png");
        final URL pause = loader.getResource("ico/pause.png");
        final URL pauseRollover = loader.getResource("ico/pauseRollover.png");
        final URL stop = loader.getResource("ico/stop.png");
        final URL stopRollover = loader.getResource("ico/stopRollover.png");
        final URL next = loader.getResource("ico/next.png");
        final URL nextRollover = loader.getResource("ico/nextRollover.png");
        final URL previous = loader.getResource("ico/previous.png");
        final URL previousRollover = loader.getResource("ico/previousRollover.png");
        final URL snap = loader.getResource("ico/snap.png");
        final URL snapRollover = loader.getResource("ico/snapRollover.png");
        final URL logos = loader.getResource("ico/logos.png");
        final URL camera = loader.getResource("ico/zoom.png");
        final URL cameraRollover = loader.getResource("ico/zoomRollover.png");
        final URL speed = loader.getResource("ico/speed.png");
        final URL speedRollover = loader.getResource("ico/speedRollover.png");

        logo.setIcon(new ImageIcon(logos));
        
        playButton.setIcon(new ImageIcon(play));
        playButton.setBorderPainted(false);
        playButton.setMargin(new Insets(0,0,0,0));
        playButton.setBorder(null);
        playButton.setText(null);
        playButton.setPressedIcon(new ImageIcon(play));
        playButton.setRolloverIcon(new ImageIcon(playRollover));

        pauseButton.setIcon(new ImageIcon(pause));
        pauseButton.setRolloverIcon(new ImageIcon(pauseRollover));
        pauseButton.setBorderPainted(false);
        pauseButton.setMargin(new Insets(0,0,0,0));
        pauseButton.setBorder(null);
        pauseButton.setText(null);

        stopButton.setIcon(new ImageIcon(stop));
        stopButton.setRolloverIcon(new ImageIcon(stopRollover));
        stopButton.setBorderPainted(false);
        stopButton.setMargin(new Insets(0,0,0,0));
        stopButton.setBorder(null);
        stopButton.setText(null);

        nextButton.setIcon(new ImageIcon(next));
        nextButton.setRolloverIcon(new ImageIcon(nextRollover));
        nextButton.setBorderPainted(false);
        nextButton.setMargin(new Insets(0,0,0,0));
        nextButton.setBorder(null);
        nextButton.setText(null);

        previousButton.setIcon(new ImageIcon(previous));
        previousButton.setRolloverIcon(new ImageIcon(previousRollover));
        previousButton.setBorderPainted(false);
        previousButton.setMargin(new Insets(0,0,0,0));
        previousButton.setBorder(null);
        previousButton.setText(null);

        snapButton.setIcon(new ImageIcon((snap)));
        snapButton.setRolloverIcon(new ImageIcon(snapRollover));
        snapButton.setBorderPainted(false);
        snapButton.setMargin(new Insets(0,0,0,0));
        snapButton.setBorder(null);
        snapButton.setText(null);

        cameraButton.setIcon(new ImageIcon((camera)));
        cameraButton.setRolloverIcon(new ImageIcon((cameraRollover)));
        cameraButton.setBorderPainted(false);
        cameraButton.setMargin(new Insets(0,0,0,0));
        cameraButton.setBorder(null);
        cameraButton.setText(null);

        speedButton.setIcon(new ImageIcon((speed)));
        speedButton.setRolloverIcon(new ImageIcon((speedRollover)));
        speedButton.setBorderPainted(false);
        speedButton.setMargin(new Insets(0,0,0,0));
        speedButton.setBorder(null);
        speedButton.setText(null);
    }
}

	