package walker;

import experiment.ExperimentState;
import experiment.Setup;
import experiment.event.capability.NodeValueEvent;
import experiment.event.capability.Directive;
import experiment.event.capability.ValueEvent;
import experiment.event.capability.LinkValueEvent;
import experiment.entity.Link;
import experiment.entity.Node;
import experiment.property.Capability;
import experiment.property.Position;
import experiment.property.Rssi;
import walker.exceptions.badSyntaxException;
import walker.exceptions.badSetupException;
import walker.exceptions.endDocumentException;
import static walker.utilities.Output.print;
import static walker.utilities.Output.print2;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import java.text.ParseException;
import java.util.Date;
import java.util.zip.GZIPInputStream;
import java.io.*;

import visual.Crawler;
import visual.Message;
import visual.entity.Entity3d;

/**
 * Created by IntelliJ IDEA.
 * User: andy
 * Date: Mar 23, 2010
 */
public class StaxParser implements XMLStreamConstants{

    private ExperimentState state;

    private String fileName;

    private ExtendedXMLStreamReader reader;

    public static boolean initialized;

    private InputStream iStream;

    // Change this to switch between relative and absolute timestamps.
    // Absolute timestamps should be unix timestamps.
    public static final boolean RELATIVE_TIMESTAMPS = true;

    // Provides more detailed print information
    public static final boolean MONITOR = false;

    // Stores the number of butes read until the beginning of the trace.
    private static int traceOffset = -1;

    private static final long SAFE_DISTANCE = 30000;

    private Message wait = new Message();

    // The jumpOffset and jumpLineOffset variables store the number of bytes and lines respectively that the previous
    // jump skipped and thus the XMLReader does not account for.
    public static long jumpOffset = 0;
    public static long  jumpLineOffset = 0; //todo: use this.

    /**
     * Default Constructor. Creates a new experiment state and parses the document.
     *
     * @param fileName the file name of the wiseml file.
     * @param state the Experiment state that is going to be associated with this parser.
     * @param timestamp the position of the file to start parsing. -1 means parse from the beginning.
     * @throws java.io.IOException if there is no such file.
     */
    public StaxParser(String fileName, ExperimentState state, double timestamp) throws IOException {

        this.fileName = fileName;
        this.state = state;

        try {
            createReader(timestamp);

            state.setXMLReader(reader);

            // Parse from the beginning of the file.
            if (timestamp == -1){

                initialized = false;
                long start;

                // Parse Setup.
                System.out.println("\n#### Parsing Setup ####");
                start = new Date().getTime();
                parseSetup();
                print("Parsing time: " + (new Date().getTime() - start) + "msec\n\n");

                // Parse Scenario.
                /*System.out.print("Parsing Scenarios.....");
                start = new Date().getTime();
                parseScenario();
                System.out.println((new Date().getTime() - start) + "msec");*/

                // Navigate to trace, even if you haven't read Scenario.
                while(!reader.check("trace")){
                    reader.nextElement();
                }

                // Save the trace offset.
                traceOffset = reader.getReader().getLocation().getCharacterOffset();

                //todo: figure what to do with traceid. Why more than one trace and how to use/visualize more.
                String traceid = reader.attr("id");
            }

            // Signal the Experiment's initialization.
            initialized = true;

        } catch(XMLStreamException e){
            e.printStackTrace();
        } catch (badSyntaxException e) {
            e.printStackTrace();
        } catch (badSetupException e) {
            e.printStackTrace();
        } catch (endDocumentException e) {
            print("Unexpected end of file.");
            e.printStackTrace();
        }


    }

    /**
     * Creates a new extended xml stream reader starting from the supplied position. Also handles gziped files.
     *
     * @param timestamp where to position the stream in the wiseml file. -1 means the beginning of the file.
     * @throws XMLStreamException if the ExtendedXMLStreamReader encounters a problem while created or preparsing.
     * @throws IOException if there is no such file.
     * @throws badSyntaxException is case of a wiseml bad syntax.
     * @throws walker.exceptions.endDocumentException if it encounters the end of the file.
     */
    public void createReader(double timestamp) throws XMLStreamException, IOException, badSyntaxException, endDocumentException {

        // If the wiseml file is gzipped,
        iStream = null;

        if (fileName.endsWith("gz")){
            // Create a stream of the gziped wiseml file.
            FileInputStream fStream = new FileInputStream(fileName);
            iStream = new GZIPInputStream(fStream);

        // If the input wiseml file isn't gzipped,
        } else {
            iStream = new FileInputStream(fileName);
        }

        // VisML Parsing from the beginning of the file.
        if (timestamp <= 0){
            reader = new ExtendedXMLStreamReader(iStream);
            jumpOffset = 0;

        // VisML Parsing from a certain point of the file.
        } else {

            // check that the seeked doesn's excheed the experiment time limits.
            if (timestamp >= Setup.getDuration()){
                throw new IOException("the given timestamp is not inside the experiment's time limits");
            }

            // Move the Stream to the seeked.
            long skipping = state.getByteOffset();
            if (skipping > traceOffset && skipping < Crawler.getFileSize()){
                skipping =  iStream.skip(skipping - SAFE_DISTANCE);
            } else {
                skipping =  iStream.skip(traceOffset - SAFE_DISTANCE);
            }
            print("Skipped to the " + skipping + " byte.");

            // Finetune the seeked by bypassing not-xml formated text
            String characters = "";
            while(!(characters.endsWith("</setup>") || characters.endsWith("</timestamp>"))){
                // each byte is translated to a character and added to the end of the characters String.
                characters = characters.concat(Character.toString((char)(iStream.read() & 0xFF)));
                if ((characters.endsWith("</wiseml>"))){
                   throw new IOException("the given position is not inside the trace limits");
                }
            }


            // Precede the InputStream with a header stream containing the necessary start tags to make the seeked
            // document an xml file.
            String header;
            if (characters.endsWith("</setup>")){
                header = "<wiseml>";
            } else {
                header = "<wiseml><trace>";
            }
            ByteArrayInputStream headerStream = new ByteArrayInputStream(header.getBytes());
            InputStream is = new SequenceInputStream(headerStream, iStream);

            // Save the number of skipped bytes.
            jumpOffset = skipping + characters.getBytes().length - header.getBytes().length;

            // Create the XML Stream Reader
            reader = new ExtendedXMLStreamReader(is);

            // Consider the state initialized.
            initialized = true;

            double time = 0;

            // In case of a clean state, avoid any parsing of timestamps.
            if (state.getTime() != 0){

                // Position the parser at the timestamp of the closer state.
                while(true){

                    if (reader.check("timestamp")){
                        if(RELATIVE_TIMESTAMPS){
                            time = Double.parseDouble(reader.elementText())*Setup.translateTime();
                        } else {
                            // absolute time in milliseconds
                            Date absoluteTime = new Date(Long.parseLong(reader.elementText())*1000);
                            time = absoluteTime.getTime() - Setup.getStart().getTime();
                        }

                        // Check for correct positioning
                        if (time >= state.getTime()){
                            if (time > state.getTime()){
                                print("Missed data! Increase safeDistance.");
                            }
                            break;
                        }
                    }
                    // Go to the next timestamp.
                    reader.navigate("timestamp");
                }
            } else {
                while (!reader.check("timestamp")){
                    reader.nextElement();
                }
                reader.nextStaxElement();
            }

            // reassign entity3ds to the seeked state's Nodes and Links
            Entity3d.rewire(state);

            // Walk until the seeked position, while updating the state.
            try {
                // display a seeking dialog
                wait.show("Skipping to seeked position", timestamp);
                // update the state to match the seeked time.
                parseTrace(time , timestamp, true);
                // hide the seeking dialog.
                wait.destroy();

            } catch (XMLStreamException e) {
                e.printStackTrace();
                // Create a file with the syntax errors that the walker found.
                Persistor.persistErrors();
            }
        }
    }

    /**
     * Main method for 'walking' a trace. It facilitate's trace parsing by mediating between the crawler and the
     * parseTrace() method. It handles positioning and range details. To avoid los of information, the
     * ExtendedXMLStreamReader should either point at a timestamp's CDATA or at a new trace element.
     *
     * @param reader the extended xml reader that keeps the stream position.
     * @param state the state to update.
     * @param speed is the number of experiment seconds per one experiment second.
     * @param fps is the number of frames per second of visualization.
     * @return true if the walker succeeded.
     * @throws badSyntaxException when a found element either not exists or it doesn't match the expectedElement.
     * @throws XMLStreamException in case of a problem with xml syntax.
     * @throws badSetupException in case of a misconfirguration in the experiment
     * @throws walker.exceptions.endDocumentException if it encounters the end of the file.
     */
    public boolean walk(ExtendedXMLStreamReader reader, ExperimentState state, double speed, double fps) throws badSyntaxException, XMLStreamException, badSetupException, endDocumentException {

        // Hold the time of the first and last timestamp to be walked.
        // They use the application's time format.
        double startTime, stopTime;

        // Don't allow walking if the experiment has not been initialized.
        if (!initialized){
            return false;
        }

        // In case of a new trace, move the reader to the first timestamp's CDATA
        if (reader.check("trace")){
            reader.navigate("timestamp");
            reader.nextStaxElement();
        // if the reader points to a timestamp element move it to the timestamp's text.
        } else if (reader.check("timestamp")){
            reader.nextStaxElement();
        }

        // Loop until you successfully read a start value that is greater than this of the experiment state.
        while(true){
            try{
                if(RELATIVE_TIMESTAMPS){
                    startTime = Double.parseDouble(reader.text())*Setup.translateTime();
                } else {
                    Date absoluteTime = new Date(Long.parseLong(reader.text())*1000);
                    startTime = absoluteTime.getTime() - Setup.getStart().getTime();
                }

                if (startTime > state.getTime()){
                    break;
                }

                reader.navigate("timestamp");
                reader.nextStaxElement();

            } catch (badSyntaxException e) {
                print("startTime couldn't be parsed, moving to the next timestamp,\n" + e.getMessage());
                reader.navigate("timestamp");
                reader.nextStaxElement();
            } catch(NumberFormatException e){
                print("startTime couldn't be parsed, due to a bad-formed character.\nMoving to the next timestamp,\n" + e.getMessage());
                reader.navigate("timestamp");
                reader.nextStaxElement();
            }
        }

        // Stop time is the last timestamp that will be parsed. In case that it is not directly set by the caller, it is
        // calculated by adding the walkingTime to state's time. WalkingTime is the duration of the walk
        // in order to visualize speed number of experiment seconds in one visualization second with fps updates of the
        // visualization per second.
        stopTime = state.getTime() + (1000*speed);

        // Check that the trace does not excheed the experiment limits.
        if (stopTime + Setup.getStart().getTime() > Setup.getEnd().getTime()){
            // if it does, report it
            print("The trace exceeds the experiment duration.");
        }

        if (stopTime < startTime){
            state.setTime(stopTime);
            return false;
        }

        // For monitoring purposes print a new line for each walk call.
         if (MONITOR) {print("");}

        // parse the trace.
        return parseTrace(startTime, stopTime, false);

    }

    /**
     * Updates the main experimentState with parsed trace values. The ExtendedXMLStreamReader should preferably point
     * at the startTime timestamp.
     *
     * @param startTime is the time of the first timestamp to be walked.
     * @param stopTime is the time of the last timestamp to be walked.
     * @param seekMode true to indicate that the method is called because of a seek action.
     * @return true if no error occured.
     * @throws XMLStreamException in case of a problem with xml syntax.
     * @throws walker.exceptions.endDocumentException if it encounters the end of the file.
     */
    private boolean parseTrace(double startTime, double stopTime, boolean seekMode) throws XMLStreamException, endDocumentException {

        // Start walking from the beginning.
        double curTime = startTime;

        // Move to the next element.
        reader.nextElement();
        
        // Main trace parsing logic.
        // Loop until stopTime is reached or a new trace is discovered.
        while ((curTime <= stopTime) && !reader.check("trace")){

            try {

                // Loop until a timestamp element is found.
                while(!reader.check("timestamp")){

                    // Construct a new empty value event.
                    ValueEvent event = null;
                    // Parse a node traceitem.
                    if (reader.check("node")){

                        String id = reader.attr("id");
                        Position position = null;

                        reader.nextElement();
                        if (reader.check("position")){
                            position = reader.position();
                        }

                        // Instantiate the new valueEvent.
                        event = new NodeValueEvent(curTime, id, position);

                        // Parse a link traceitem.
                    } else if (reader.check("link")){

                        String sourceId = reader.attr("source");
                        String targetId = reader.attr("target");
                        Rssi rssi = null;
                        reader.nextElement();

                        if(reader.check("rssi")){
                            rssi = reader.rssi();
                            reader.nextElement();
                        }

                        // Instantiate the new valueEvent.
                        event = new LinkValueEvent(curTime, sourceId, targetId, rssi);

                    // If a trace element is encountered, return.
                    } else if (reader.check("trace")){
                        return true;
                    }

                    // Check that the event has been created.
                    if (event == null){
                        throw new badSyntaxException("could not costruct a new Event");
                    }

                    // For each data element add a Directive to the ValueEvent.
                    while(reader.check("data")){

                        Directive directive = null;
                        try {
                            directive = new Directive(reader.attr("key"), reader.elementText());
                        } catch (badSyntaxException e){
                            Persistor.error(reader.getLocation());

                        }

                        // If this a valid directive, add it.
                        if (directive != null){
                            event.addData(directive);
                        }

                        // move to the next element.
                        reader.nextElement();
                    }

                    // If possible apply the new ValueEvent to the given experiment state.
                    if (!event.apply(state)){
                        throw new badSyntaxException("ValueEvent could not be applied");
                    }

                    // In case that the walk was called from a seek action,
                    if (seekMode){
                        // update the seek progressBar
                        wait.setProgress(curTime);

                    // If it was called from Crawler and in monitor mode,
                    } else if (MONITOR){
                        // monitor each update.
                        print2("*");
                    }

                }

            } catch (badSyntaxException e){
                print("mistake in walking,");
                e.printStackTrace();

                // try to recover
                try {
                    while(!reader.check("timestamp")){
                        reader.nextElement();
                    }
                } catch (XMLStreamException e1) {
                    print("could not recover. The next timestamp cannot be found. Something is going wrong!!");
                    e1.printStackTrace();
                    return false;
                } catch (endDocumentException e2){
                    print("could not recover. The next timestamp cannot be found. \nThe end of the file was reached. " +
                            "\nSomething is going wrong!!\n");
                    e2.printStackTrace();
                    return false;
                }
            }

            // VisML parsing the following timestamp.
            try {
                // Update current time.
                if (reader.check("timestamp")){
                    if(RELATIVE_TIMESTAMPS){
                        curTime = Double.parseDouble(reader.elementText())*Setup.translateTime();
                    } else {
                        // absolute time in milliseconds
                        Date absoluteTime = new Date(Long.parseLong(reader.elementText())*1000);
                        curTime = absoluteTime.getTime() - Setup.getStart().getTime();
                    }

                    // Don't move to the next element if this was the last loop.
                    if (curTime <= stopTime){
                        // Move to the next element.
                        reader.nextElement();
                    }
                }

            } catch (badSyntaxException e) {
                print("Unable to read the following timestamp");
            }

        }

        // if the user seeked,
        if (seekMode){
            // update the state's time.
            try {
                state.setTime(stopTime);
            } catch (badSyntaxException e) {
                e.printStackTrace();
            }
        }
        return true;
    }

    /**
     * Parses the setup information of the experiment while checking for correct syntax.
     * 
     * @throws badSyntaxException if a found element either not exists or it doesn't match the expectedElement
     * @throws XMLStreamException in case of a problem with xml syntax.
     * @throws badSetupException in case of a misconfirguration in the experiment
     * @throws walker.exceptions.endDocumentException if it encounters the end of the file.
     */
    private void parseSetup() throws badSyntaxException, XMLStreamException, badSetupException, endDocumentException {

        reader.navigate("setup");
        reader.nextElement();

        // Parsing origin Position
        if(reader.check("origin")){
            //parse it and inform Setup
            Setup.setOrigin(reader.position());
        } else {
            // initialize Setup with default values
            Setup.setOrigin(new Position(0,0,0,0,0));
        }
        print("   ..origin parsed");

        // Parsing time information
        if (reader.check("timeinfo")){
            Date start;
            try{
                start = Setup.dateFormat.parse(reader.navigateRead("start"));
            } catch (ParseException e) {
                throw new badSyntaxException("wrong date");
            }
            reader.nextElement();

            if (reader.check("end")){
                Date end;
                try{
                    end = Setup.dateFormat.parse(reader.elementText());
                } catch (ParseException e) {
                    throw new badSyntaxException("wrong date");
                }
                Setup.Unit unit = Setup.toUnit(reader.navigateRead("unit"));

                Setup.setTime(start, end, unit);

            } else if (reader.check("duration")){
                long duration = Long.parseLong(reader.elementText());
                Setup.Unit unit = Setup.toUnit(reader.navigateRead("unit"));

                Setup.setTime(start, duration, unit);

            } else {
                throw new badSyntaxException("(duration||end-timestamp) of the experiment is required.");
            }
            print("   ..timeinfo parsed");
            reader.nextElement();

        } else {
            print("..no time info. Starting time is considered to be '0'.");
        }

        // Parsing other experiment information
        if (reader.check("interpolation")){
            Setup.setInterpolation(Setup.Interpolation.valueOf(reader.elementText()));
            print("   ..interpolation set");
            reader.nextElement();
        } else {
            // todo set default interpolation
        }

        if(reader.check("coordinateType")){
            Setup.setCoordType(reader.elementText());
            print("   ..coordinateType set");
            reader.nextElement();
        } else {
            Setup.setCoordType(Setup.CoordinateType.cartesian);
        }

        if (reader.check("description")){
            Setup.setDescription(reader.elementText());
            print("   ..description set");
            reader.nextElement();
        }

        // Parse default values.
        if (reader.check("defaults")){
            print("Setting default entity values:");
            reader.nextElement();

            if (reader.check("node")){
                print("   ..parsing default Node");
                node(true);
            }

            if (reader.check("link")){
                print("   ..parsing default Link");
                link(true);
            }
        }

        // Parsing Nodes
        if(!reader.check("node")){
            throw new badSyntaxException("There should be at least one node element");
        }

        do {
            node(false);
        } while (reader.check("node"));

        print("Number of experiment Nodes: " + state.getNodeNumber());

        // Parsing Links
        while (reader.check("link")) {
            link(false);
        }
        print("Number of experiment Links: " + state.getLinkNumber());




    }

    /**
     * Reads a single Link element and moves the stream to the following element.
     *
     * @param defaultMode true only for parsing the defaultNode, false for parsing topology
     * @throws badSyntaxException if the wiseml file has bad syntax
     * @throws XMLStreamException if there is a problem with xml syntax.
     * @throws badSetupException in case of a misconfirguration in the experiment.
     * @throws walker.exceptions.endDocumentException if it encounters the end of the file.
     */
    protected  void link(boolean defaultMode) throws badSyntaxException, XMLStreamException, badSetupException, endDocumentException {

        Link link = null;
        Node source = null, target = null;

        if(!defaultMode){
            source = state.indexNode(reader.attr("source"));
            target = state.indexNode(reader.attr("target"));

            // Check not found case.
            if(source == null || target == null){
                throw new badSetupException("Could not index link attributes");
            }
        }


        String encrypted = null, virtual = null;
        Rssi rssi = null;

        reader.nextElement();

        // Parsing Link encrypted value
        if (reader.check("encrypted")){
            encrypted = reader.elementText();
            boolean enc;
            try{
                enc = Boolean.parseBoolean(encrypted);
            } catch(IllegalArgumentException e){
                throw new badSyntaxException("Illegal type in <encrypted>");
            }
            if (defaultMode){
                    Link.defaultIsEncrypted = enc;
                }
            reader.nextElement();
        }

        // Parsing default Link isVirtual value
        if (reader.check("virtual")){
            virtual = reader.elementText();
            boolean isV;
            try{
                isV = Boolean.parseBoolean(virtual);
            } catch(IllegalArgumentException e){
                throw new badSyntaxException("Illegal type in <virtual>");
            }
            if (defaultMode){
                Link.defaultIsVirtual = isV;
            }
            reader.nextElement();
        }

        // Parsing default Link RSSI value
        if (reader.check("rssi")){
            rssi = reader.rssi();
            if(defaultMode){
                Link.defaultRssi = rssi;
            }
            reader.nextElement();
        }

        if (!defaultMode){
            link = new Link(state, source, target, encrypted, virtual, rssi);
        }

        // Parsing default Node3d capabilities
        while (reader.check("capability")){
            Capability c = reader.capability();
            if(!defaultMode){
                link.setCapability(c);
            } else {
                Link.addDefaultCapability(c);
            }
            reader.nextElement();
        }


    }

    /** Reads a single node element and moves the stream to the following element.
     *
     * @param defaultMode true only for parsing the defaultNode, false for parsing topology
     * @throws badSyntaxException if the wiseml file has bad syntax
     * @throws XMLStreamException if there is a problem with xml syntax.
     * @throws walker.exceptions.endDocumentException if it encounters the end of the file.
     */
    protected void node(boolean defaultMode) throws badSyntaxException, XMLStreamException, endDocumentException {

        Node node = null;
        String id = null;
        if (!defaultMode){
            id = reader.attr("id");
        }
        Position position = null;
        String gateway = null;
        String programDetails = null, nodeType = null, description = null;

        reader.nextElement();

        // Parsing seeked element.
        if (reader.check("position")){
            position = reader.position();
            if(defaultMode){
                Node.defaultPosition = position;
            }
        }

        // Parsing gateway element.
        if (reader.check("gateway")){
            gateway = reader.elementText();
            // check if <gateway> has the correct value
            boolean gw;
            try{
                gw = Boolean.parseBoolean(gateway);
            } catch(IllegalArgumentException e){
                throw new badSyntaxException("Illegal type in <gateway>");
            }
            if(defaultMode){
                Node.defaultGateway = gw;
            }
            reader.nextElement();
        }

        // Parsing programDetails element.
        if (reader.check("programDetails")){
            programDetails = reader.elementText();
            if(defaultMode){
                Node.defaultProgDetails = programDetails;
            }
            reader.nextElement();
        }

        // Parsing nodeType element.
        if (reader.check("nodeType")){
            nodeType = reader.elementText();
            if(defaultMode){
                Node.defaultNodeType = nodeType;
            }
            reader.nextElement();
        }

        // Parsing description element.
        if (reader.check("description")){
            description = reader.elementText();
            if(defaultMode){
                Node.defaultDescription = description;
            }
            reader.nextElement();
        }

        // Add the Node
        if(!defaultMode){
            node = new Node(state, id, position, gateway, programDetails, nodeType, description );
        }

        // Parsing default Node capabilities
        while (reader.check("capability")){

            Capability c = reader.capability();
            if(!defaultMode){
                node.setCapability(c);
            } else {
                Node.addDefaultCapability(c);
            }
            reader.nextElement();
        }

    }

    /**
     * @return the wiseml fileName
     */
    public String getFileName() {
        return fileName;
    }

    /**
     * @return StaxParser's experiment State.
     */
    public ExperimentState getState() {
        return state;
    }

    /**
     * @return the ExtendedXMLStreamReader that is associated with StaxParser.
     */
    public ExtendedXMLStreamReader getReader() {
        return reader;
    }

    /**
     * Creates a new extended xml stream reader starting from the supplied position. Also handles gziped files.
     *
     * @param position where to position the stream in the wiseml file. -1 means the beginning of the file.
     * @throws XMLStreamException if the ExtendedXMLStreamReader encounters a problem while created or preparsing.
     * @throws IOException if there is no such file.
     * @throws badSyntaxException is case of a wiseml bad syntax.
     */
    public void createReaderOld(long position) throws XMLStreamException, IOException, badSyntaxException {

        // If the wiseml file is gzipped,
        iStream = null;
        if (fileName.endsWith("gz")){
            // Create a stream of the gziped wiseml file.
            FileInputStream fStream = new FileInputStream(fileName);
            iStream = new GZIPInputStream(fStream);

            // If the input wiseml file isn't gzipped,
        } else {
            iStream = new FileInputStream(fileName);
        }

        // VisML Parsing from the beginning of the file.
        if (position == -1){
            reader = new ExtendedXMLStreamReader(iStream);

            // VisML Parsing from a seeked in the file.
        } else {
            // check that the seeked doesn's excheed the file length.
            if (position >= Crawler.getFileSize()){
                throw new IOException("the given position is not inside the file limits");
            }

            // Move the Stream to the seeked.
            print("Skipped to the " + iStream.skip(position) + " byte.");

            // Finetune the seeked by bypassing not-xml formated text
            String characters = "";
            while(!(characters.endsWith("</setup>") || characters.endsWith("</timestamp>"))){
                // each byte is translated to a character and added to the end of the characters String.
                characters = characters.concat(Character.toString((char)(iStream.read() & 0xFF)));
                if ((characters.endsWith("</wiseml>"))){
                    throw new IOException("the given position is not inside the trace limits");
                }
            }

            // Precede the InputStream with a header stream containing the necessary start tags to make the seeked
            // document an xml file.
            String header;
            if (characters.endsWith("</setup>")){
                header = "<wiseml>";
            } else {
                header = "<wiseml><trace>";
            }
            ByteArrayInputStream headerStream = new ByteArrayInputStream(header.getBytes());
            InputStream is = new SequenceInputStream(headerStream, iStream);

            // Create the XML Stream Reader
            reader = new ExtendedXMLStreamReader(is);

            /*int k = 0;
            while(k++<30){
                reader.nextStaxElement();
                reader.debug();
            }*/
        }
    }

}
